/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Implementations of the actions. */

/* The general theory of actions is that the interface or AI code calls, for
   an action foo, the routine prep_foo_action, which just records the action
   for later execution.  The action loop in run_game eventually calls
   do_foo_action, which first calls check_foo_action to confirm that the
   action will succeed.  If check_foo_action does not find any errors, then
   the action cannot fail.  The main body of do_foo_action then implements
   the effects of the action.  Interfaces may call check_foo_action freely,
   but should never call do_foo_action directly. */

#include "conq.h"

/* The table of all the types of actions. */

ActionDefn actiondefns[] = {

#undef  DEF_ACTION
#define DEF_ACTION(CODE,NAME,ARGS,FN,doc) { CODE, NAME, ARGS, FN },

#include "action.def"

    { -1, NULL, NULL, NULL }
};

char actiondesigbuf[BUFSIZE];

/* Eventually this should do initial alloc of actions. */

init_actions()
{
}

/* Just a placeholder action, so not much to do here. */

do_none_action(unit)
Unit *unit;
{
    return A_ANY_DONE;
}

/* (should stash unit ids instead of ptrs) */

prep_move_action(unit, x, y, z, unit2)
Unit *unit, *unit2;
int x, y, z;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
    unit->act->nextaction.type = A_MOVE;
    unit->act->nextaction.args[0] = x;
    unit->act->nextaction.args[1] = y;
    unit->act->nextaction.args[2] = z;
    unit->act->nextaction.args[3] = (long) unit2;
    return TRUE;
}

/* The basic act of moving.  This attempts to go and maybe fails, but
   takes no corrective action.  Note that this requires space in the
   destination hex, will not board, attack, etc - all that is task- and
   plan-level behavior. */

do_move_action(unit, nx, ny, nz, unit2)
Unit *unit, *unit2;
int nx, ny, nz;
{
    int u, u2, mpcost = 0, acpcost;
    int rslt = check_move_action(unit, nx, ny, nz, unit2);

    if (!valid(rslt)) return rslt;
    u = unit->type;
    u2 = unit2->type;
    if (!inside_area(nx, ny)) {
	kill_unit(unit2, LEFTMAP);
	rslt = A_ANY_DONE;
    } else if (ut_vanishes_on(u2, terrain_at(nx, ny))) {
	kill_unit(unit2, DISASTER);
	rslt = A_ANY_DONE;
    } else if (ut_wrecks_on(u2, terrain_at(nx, ny))) {
	kill_unit(unit2, DISASTER);
	rslt = A_ANY_DONE;
    } else {
	mpcost = move_unit(unit2, nx, ny);
	rslt = A_ANY_DONE;
    }
    acpcost = u_acp_to_move(u);
    if (u_acp_per_mp(u) > 0) {
	if (u_acp_per_mp(u) == 100) {
	    acpcost = mpcost;
	} else {
	    acpcost += (u_acp_per_mp(u) * mpcost) / 100;
	}
    }
    use_up_acp(unit, acpcost); 
    return rslt;
}

check_move_action(unit, nx, ny, nz, unit2)
Unit *unit, *unit2;
int nx, ny, nz;
{
    int u, u2, ox, oy, oz, mpavail = 0, totcost, freemp, m;

    if (!in_play(unit)) return A_ANY_ERROR;
    /* Note that edge cell dests (used to leave the world) will be allowed. */
    if (!in_area(nx, ny)) return A_ANY_ERROR;
    if (!in_play(unit2)) return A_ANY_ERROR;
    u = unit->type;
    u2 = unit2->type;
    ox = unit2->x;  oy = unit2->y;  oz = unit2->z;
    if (u_acp_to_move(u) < 1) return A_ANY_CANNOT_DO;
    if (!has_enough_acp(unit, u_acp_to_move(u))) return A_ANY_NO_ACP;
    if (!has_supply_to_act(unit)) return A_ANY_NO_SUPPLY;
    /* Destination is outside the world and we're not allowed to leave. */
    if (!inside_area(nx, ny) && u_mp_to_leave_world(u2) < 0) {
	return A_MOVE_CANNOT_LEAVE_WORLD;
    }
    /* Check if the destination is within our move range. */
    /* (but check for border slides here?) */
    if (distance(ox, oy, nx, ny) > u_move_range(u2)) return A_MOVE_TOO_FAR;
    if (nz > 0) return A_MOVE_TOO_FAR;
    /* Now start looking at the move costs. */
    totcost = total_move_cost(u2, ox, oy, oz, nx, ny, nz);
    if (unit2->act) mpavail = unit2->act->mp;
/*    if (u_speed(u2) < totcost) return A_ANY_ERROR; why was this here? */
    if (mpavail <= 0) return A_MOVE_NO_MP;
    /* The free mp might get us enough moves, so add it before comparing. */
    if (mpavail + u_free_mp(u2) < totcost) return A_MOVE_NO_MP;
    /* If cell terrain is too small, we can't move into it. */
    if (!can_occupy_cell(unit2, nx, ny)) return A_MOVE_DEST_FULL;
    /* We have to have a minimum level of supply to be able to move. */
    for_all_material_types(m) {
	if (unit->supply[m] < um_to_move(u, m)) return A_ANY_NO_SUPPLY;
    }
    return A_ANY_OK;
}

/* Conduct the actual move (used in both normal moves and some combat). */

move_unit(unit, nx, ny)
Unit *unit;
int nx, ny;
{
    int ox = unit->x, oy = unit->y, mpcost = 0;

    /* Disappear from the old location ... */
    leave_hex(unit);
    /* ... and appear in the new one! */
    enter_hex(unit, nx, ny);
    /* Movement may set off other people's alarms. */
    maybe_react_to_move(unit, ox, oy);
    /* Count against movement points. */
    if (unit->act) {
	mpcost = total_move_cost(unit->type, ox, oy, 0, nx, ny, 0);
	unit->act->mp -= mpcost;
	unit->act->actualmoves += 1;
    }
    /* Use up supplies as directed. */
    consume_move_supplies(unit);
    return mpcost;
}

can_move_at_all(unit)
Unit *unit;
{
    return u_speed(unit->type) > 0;
}

/* Compute the number of move points that will be needed to do the given
   move. */

total_move_cost(u, x1, y1, z1, x2, y2, z2)
int u, x1, y1, z1, x2, y2, z2;
{
    int cost, b, c, conncost;

    if (z1 != 0 || z2 != 0) {
    	/* should write these calcs eventually */
    }
    /* are the two points adjacent? */
    cost = ut_mp_to_leave(u, terrain_at(x1, y1));
    if (numbordtypes > 0) {
	    /* Add any necessary border crossing costs. */
	    for_all_terrain_types(b) {
		if (t_is_border(b)
		    && aux_terrain_defined(b)
		    && border_at(x1, y1, closest_dir(x2 - x1, y2 - y1), b)) {
		    cost += ut_mp_to_enter(u, b);
		}
	    }
    }
    cost += ut_mp_to_enter(u, terrain_at(x2, y2));
    /* Use a connection traversal if it would be cheaper.  This is only automatic
       if the connection on/off costs are small enough, otherwise the unit has
       to do explicit actions to get on the connection and off again. */
    if (numconntypes > 0) {
	    /* Try each connection type to see if it's better. */
	    for_all_terrain_types(c) {
		if (t_is_connection(c)
		    && aux_terrain_defined(c)
		    && connection_at(x1, y1, closest_dir(x2 - x1, y2 - y1), c)) {
		    conncost = ut_mp_to_enter(u, c)
		    	       + ut_mp_to_traverse(u, c)
		    	       + ut_mp_to_leave(u, c);
		    cost = min(cost, conncost);
		}
	    }
    }
    /* The cost of leaving the world is always an addon. */
    if (!inside_area(x2, y2)) {
    	cost += u_mp_to_leave_world(u);
    }
    return max(1, cost);
}

closest_dir(x, y)
int x, y;
{
    int dir;

    for_all_directions(dir) {
	if (dirx[dir] == x && diry[dir] == y) return dir;
    }
    return 0;
}

/* This is a hook to handle any reactions to the unit's successful move. */

maybe_react_to_move(unit, ox, oy)
Unit *unit;
int ox, oy;
{
}

consume_move_supplies(unit)
Unit *unit;
{
    int u = unit->type, m;
    
    for_all_material_types(m) {
	if (um_consume_per_move(u, m) > 0) {
	    unit->supply[m] -= um_consume_per_move(u, m);
#if 0
	    if (unit->supply[m] <= 0 && unit->transport == NULL) {
		if (um_consume(u, m) > 0) {
		    exhaust_supply(unit); /* don't always check right away? */
		    return;
		}
	    }
#endif
	}
    }
}

/* Movement into another unit. */

prep_enter_action(unit, dest, occ)
Unit *unit, *dest, *occ;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
    unit->act->nextaction.type = A_ENTER;
    unit->act->nextaction.args[0] = (long) dest;
    unit->act->nextaction.args[1] = (long) occ;
    return TRUE;
}

do_enter_action(unit, dest, occ)
Unit *unit, *dest, *occ;
{
    int rslt = check_enter_action(unit, dest, occ);

    if (!valid(rslt)) return rslt;
    leave_hex(occ);
    enter_transport(occ, dest);
    /* this is doctrine? */
    /*	if (n > 0) order_sentry(unit, n); */
    return A_ANY_DONE;
}

check_enter_action(unit, dest, occ)
Unit *unit, *dest, *occ;
{
    int u, u2;

    if (!in_play(unit)) return A_ANY_ERROR;
    if (!in_play(occ)) return A_ANY_ERROR;
    if (!in_play(dest)) return A_ANY_ERROR;
    u = unit->type;
    u2 = occ->type;
    /* Moving self or another? */
    if (unit != occ) {
	/* under what conditions is this permissible? */
	return A_ANY_ERROR;
    }
    if (dest == NULL) return A_ANY_ERROR;
    if (!can_occupy(occ, dest)) return A_ANY_ERROR;
    if (!has_enough_acp(unit, uu_acp_to_enter(u, u2))) return A_ANY_NO_ACP;
    return A_ANY_OK;
}

/* Explicit material production. */

do_produce_action(unit, r, n, unit2)
Unit *unit, *unit2;
int r, n;
{
    int rslt = check_produce_action(unit, r, n, unit2);

    if (!valid(rslt)) return rslt;
    use_up_acp(unit, 1);
    return A_ANY_DONE;
}

check_produce_action(unit, r, n, unit2)
Unit *unit, *unit2;
int r, n;
{
    if (!has_enough_acp(unit, 1)) return A_ANY_NO_ACP;
    return A_ANY_ERROR;
}

/* Transfer of material. */

do_transfer_action(unit, r, n, unitfrom, unitto)
Unit *unit, *unitfrom, *unitto;
int r, n;
{
    int actual;
    int rslt = check_transfer_action(unit, r, n, unitfrom, unitto);

    if (!valid(rslt)) return rslt;
    actual = transfer_supply(unitfrom, unitto, r, n);
    use_up_acp(unit, 1);
    if (actual == n) {
	return A_ANY_DONE;
    } else {
	return A_ANY_DONE;
    }
}

check_transfer_action(unit, r, n, unitfrom, unitto)
Unit *unit, *unitfrom, *unitto;
int r, n;
{
    if (!has_enough_acp(unit, 1)) return A_ANY_NO_ACP;
    return A_ANY_ERROR;
}

/* Move supply from one unit to another.  Don't move more than is possible; */
/* check both from and to amounts and capacities. */

transfer_supply(from, to, r, amount)
Unit *from, *to;
int r, amount;
{
    amount = min(amount, from->supply[r]);
    amount = min(amount, um_storage(to->type, r) - to->supply[r]);
    from->supply[r] -= amount;
    to->supply[r] += amount;
    Dprintf("%s receives %d %s\n", unit_desig(to), amount, m_type_name(r));
    return amount;
}

/* If a side's tech level is under its max, research can increase it. */

prep_research_action(unit, u2)
Unit *unit;
int u2;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
/*    unit->act->nextaction.type = A_CREATE_IN; */
    unit->act->nextaction.args[0] = u2;
    return TRUE;
}

do_research_action(unit, u2)
Unit *unit;
int u2;
{
    int u = unit->type;
    int rslt = check_research_action(unit, u2);

    if (!valid(rslt)) return rslt;
    unit->side->tech[u2] += prob_fraction(uu_tech_per_research(u, u2));
#if 0
    /* Silently apply the per-side-per-turn limit on tech gains. */
    unit->side->tech[u2] =
	min(unit->side->tech[u2], unit->side->inittech[u2] + maxgain);
#endif
    /* Adjust the tech levels of any related unit types to match. */
    adjust_tech_crossover(unit->side, u2);
    /* (should notify side about changes and/or thresholds reached?) */
    use_up_acp(unit, uu_acp_to_research(u, u2));
    return A_ANY_DONE;
}

prob_fraction(n)
int n;
{
	return (n / 100 + (probability(n % 100) ? 1 : 0));
}

check_research_action(unit, u2)
Unit *unit;
int u2;
{
    int u;
    Side *side;

    if (!in_play(unit)) return A_ANY_ERROR;
    if (!is_unit_type(u2)) return A_ANY_ERROR;
    u = unit->type;
    side = unit->side;
    /* Independent units don't do research. */
    if (side == NULL) return A_ANY_ERROR;
    /* This unit must be of a type that can research the given type. */
    if (uu_acp_to_research(u, u2) < 1) return A_ANY_CANNOT_DO;
    /* Max tech level means there's nothing more to learn. */
    if (side->tech[u2] >= u_tech_max(u2)) return A_ANY_ERROR;
    if (!has_enough_acp(unit, uu_acp_to_research(u, u2))) return A_ANY_NO_ACP;
    return A_ANY_OK;
}

/* For all unit types, bring their tech level up to match the crossovers
   from the given unit type. */

adjust_tech_crossover(side, u)
Side *side;
int u;
{
    int u2, cross;

    for_all_unit_types(u2) {
	if (u2 != u) {
	    /* (should be "as ratio of max levels for each type") */
	    cross = (uu_tech_crossover(u, u2) * side->tech[u2]) / 100;
	    if (cross > side->tech[u2]) side->tech[u2] = cross;
	}
    }
}

/* Before a unit can build another, it must take some time to prepare by
   "tooling up". */

prep_toolup_action(unit, u2)
Unit *unit;
int u2;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
/*    unit->act->nextaction.type = A_CREATE_IN; */
    unit->act->nextaction.args[0] = u2;
    return TRUE;
}

do_toolup_action(unit, u2)
Unit *unit;
int u2;
{
    int rslt = check_toolup_action(unit, u2);

    if (!valid(rslt)) return rslt;
    use_up_acp(unit, uu_acp_to_toolup(unit->type, u2));
    return A_ANY_DONE;
}

check_toolup_action(unit, u2)
Unit *unit;
int u2;
{
    if (!in_play(unit)) return A_ANY_ERROR;
    if (!is_unit_type(u2)) return A_ANY_ERROR;
    if (uu_acp_to_toolup(unit->type, u2) < 1) return A_ANY_CANNOT_DO;
    if (!has_enough_acp(unit, uu_acp_to_toolup(unit->type, u2)))
      return A_ANY_NO_ACP;
    return A_ANY_OK;
}

prep_create_in_action(unit, u2, transport)
Unit *unit, *transport;
int u2;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
    unit->act->nextaction.type = A_CREATE_IN;
    unit->act->nextaction.args[0] = u2;
    unit->act->nextaction.args[1] = (long) transport;
    return TRUE;
}

/* (how do we say that the new unit is to be built *around* some other
   unit(s)? ) */

/* This action creates the (incomplete) unit. */

do_create_in_action(unit, u2, transport)
Unit *unit, *transport;
int u2;
{
    int u;
    Unit *unit2;
    int rslt = check_create_in_action(unit, u2, transport);

    if (!valid(rslt)) return rslt;
    u = unit->type;
    /* Make the new unit. */
    if ((unit2 = create_unit(u2)) != NULL) {
	/* Flag it as incomplete. */
	unit2->cp = uu_creation_cp(u, u2);
	unit2->hp = 1;
	if (unit2->cp >= u_cp(u2)) {
	    make_unit_complete(unit2);
	}
	/* (why change side here?) */
	set_unit_side(unit2, unit->side);
	enter_transport(unit2, transport);
	use_up_acp(unit, uu_acp_to_create(u, u2));
	return A_ANY_DONE;
    } else {
	/* We've hit a max number of units, nothing to be done. */
	return A_ANY_ERROR;
    }
}

check_create_in_action(unit, u2, transport)
Unit *unit, *transport;
int u2;
{
    int u;

    if (!in_play(unit)) return A_ANY_ERROR;
    if (!is_unit_type(u2)) return A_ANY_ERROR;
    if (!in_play(transport)) return A_ANY_ERROR;
    u = unit->type;
    if (uu_acp_to_create(u, u2) < 1) return A_ANY_CANNOT_DO;
    /* Check the tech level of the side. */
    if (u_tech_to_build(u2) > 0) {
	if (unit->side == NULL) return A_ANY_ERROR;
	if (unit->side->tech[u2] < u_tech_to_build(u2)) return A_ANY_ERROR;
    }
    /* (check tooling also) */
    if (!type_can_occupy(u2, transport)) return A_ANY_ERROR;
    if (!has_enough_acp(unit, uu_acp_to_create(u, u2))) return A_ANY_NO_ACP;
    return A_ANY_OK;
}

prep_create_at_action(unit, u2, x, y, z)
Unit *unit;
int u2, x, y, z;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
    unit->act->nextaction.type = A_CREATE_IN;
    unit->act->nextaction.args[0] = u2;
    unit->act->nextaction.args[1] = x;
    unit->act->nextaction.args[2] = y;
    unit->act->nextaction.args[3] = z;
    return TRUE;
}

do_create_at_action(unit, u2, x, y, z)
Unit *unit;
int u2, x, y, z;
{
    int u;
    Unit *unit2;
    int rslt = check_create_at_action(unit, u2, x, y, z);

    if (!valid(rslt)) return rslt;
    u = unit->type;
    /* Make the new unit. */
    if ((unit2 = create_unit(u2)) != NULL) {
	unit2->cp = uu_creation_cp(u, u2);
	unit2->hp = 1;
	if (completed(unit2)) {
	    make_unit_complete(unit2);
	}
	set_unit_side(unit2, unit->side);
	if (0 /* builder should go inside */) {
	    /* take the builder off the hex and put it inside the new unit. */
	}
	enter_hex(unit2, x, y);
	/* and set its altitude? */
	use_up_acp(unit, uu_acp_to_create(u, u2));
	return A_ANY_DONE;
    } else {
	/* We've hit a max number of units, nothing to be done. */
	return A_ANY_ERROR;
    }
}

check_create_at_action(unit, u2, x, y, z)
Unit *unit;
int u2, x, y, z;
{
    int u;

    if (!in_play(unit)) return A_ANY_ERROR;
    if (!is_unit_type(u2)) return A_ANY_ERROR;
    if (!inside_area(x, y)) return A_ANY_ERROR;
    u = unit->type;
    if (uu_acp_to_create(u, u2) < 1) return A_ANY_CANNOT_DO;
    if (u_tech_to_build(u2) > 0) {
	if (unit->side == NULL) return A_ANY_ERROR;
	if (unit->side->tech[u2] < u_tech_to_build(u2)) return A_ANY_ERROR;
    }
    if (distance(unit->x, unit->y, x, y) > uu_sep_to_create_max(u, u2))
      return A_ANY_ERROR;
    /* check tooling */
    /* (should check for room and safety of terrain) */
    if (!type_can_occupy_cell(u2, x, y)) return A_ANY_ERROR;
    /* (should check that unit limit not hit yet) */
    if (!has_enough_acp(unit, uu_acp_to_create(u, u2))) return A_ANY_NO_ACP;
    return A_ANY_OK;
}

prep_build_action(unit, unit2)
Unit *unit, *unit2;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
    unit->act->nextaction.type = A_BUILD;
    unit->act->nextaction.args[0] = (long) unit2;
    return TRUE;
}

/* This action makes progress on a construction effort, possibly completing
   the new unit and making it available. */

do_build_action(unit, unit2)
Unit *unit, *unit2;
{
    int u, u2, r, mk, rmk, use;
    Side *us;
    int rslt = check_build_action(unit, unit2);

    if (!valid(rslt)) return rslt;
    u = unit->type;
    u2 = unit2->type;
#if 0
    /* Consume materials. */
    mk = uu_make(u, unit->product);
    for_all_material_types(r) {
	if ((rmk = um_tomake(unit->product, r)) > 0) {
	    use = (rmk / mk) + (unit->schedule <= (rmk % mk) ? 1 : 0);
	    if (unit->supply[r] < use) return;
	    unit->supply[r] -= use;
	}
    }
#endif
    unit2->cp += uu_cp_per_build(u, u2);
    if (completed(unit2)) {
	make_unit_complete(unit2);
	/* Get rid of the building unit if it is to be the garrison. */
	if (will_garrison(u, u2)) {
	    /* But first get the about-to-be-killed garrisoning unit
	       disconnected from everything. */
	    /* (still need to do something with other occs...) */
	    /* (how about moving to various places, killing them
	       only as absolute last resort) */
	    if (unit2->transport == unit) {
		leave_transport(unit2);
		if (unit->transport != NULL) {
		    enter_transport(unit2, unit->transport);
		} else {
		    enter_hex(unit2, unit->x, unit->y);
		}
	    }
	    /* Now we can get rid of the garrison w/o scrambling anything. */
	    kill_unit(unit, GARRISON);
	}
    }
    update_unit_display(unit2->side, unit2, TRUE);
    use_up_acp(unit, uu_acp_to_build(u, u2));
    return A_ANY_DONE;
}

check_build_action(unit, unit2)
Unit *unit, *unit2;
{
    int u, u2;

    if (!in_play(unit)) return A_ANY_ERROR;
    if (!in_play(unit2)) return A_ANY_ERROR;
    u = unit->type;
    u2 = unit2->type;
    /* Can't finish building a unit until we have the technology. */
    if (u_tech_to_build(u2) > 0) {
	if (unit->side == NULL) return A_ANY_ERROR;
	if (unit->side->tech[u2] < u_tech_to_build(u2)) return A_ANY_ERROR;
    }
    if (distance(unit->x, unit->y, unit2->x, unit2->y)
	> uu_sep_to_build_max(u, u2))
      return A_ANY_ERROR;
    if (!has_enough_acp(unit, uu_acp_to_build(u, u2))) return A_ANY_NO_ACP;
    return A_ANY_OK;
}

make_unit_complete(unit)
Unit *unit;
{
    int u = unit->type;

    /* Make this a full-fledged unit. */
    unit->cp = u_cp(u) /* / u_parts(u) ? */;
    unit->hp = u_hp(u);
    /* Christen our new unit. */
    make_up_unit_name(unit);
    Dprintf("%s is completed\n", unit_desig(unit));
    /* Count this new unit in tallies. */
    /*	(side_balance(us, u2, PRODUCED))++;  */
    /*	(us->units[u2])++;  */
}

/* Repair action. */

prep_repair_action(unit, unit2)
Unit *unit, *unit2;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
    unit->act->nextaction.type = A_REPAIR;
    unit->act->nextaction.args[0] = (long) unit2;
    return TRUE;
}

do_repair_action(unit, unit2)
Unit *unit, *unit2;
{
    int u, u2, r;
    int rslt = check_repair_action(unit, unit2);

    if (!valid(rslt)) return rslt;
    u = unit->type;
    u2 = unit2->type;
#if 0
    if ((r = random(uu_repair(u,u2))) == 0) {
	for_all_material_types(r) {
	    if (unit->supply[r] < um_tomake(u2, r)) {
		Dprintf("Unit not repaired for lack of supplies");
		return;
	    }
	}
	for_all_material_types(r) {
	    unit->supply[r] -= um_tomake(u2, r) / u_hp(u2);
	}
    }
#endif
    /* Add to the repairee's hit points. */
    unit2->hp++;
    use_up_acp(unit, uu_acp_to_repair(u, u2));
    return A_ANY_DONE;
}

check_repair_action(unit, unit2)
Unit *unit, *unit2;
{
    int u, u2, m;

    if (!in_play(unit)) return A_ANY_ERROR;
    if (!in_play(unit2)) return A_ANY_ERROR;
    u = unit->type;
    u2 = unit2->type;
    if (uu_acp_to_repair(u, u2) < 1) return A_ANY_CANNOT_DO;
    if (uu_repair(u, u2) <= 0) return A_ANY_ERROR;
    if (unit2->hp >= u_hp(u2)) return A_ANY_ERROR;
    if (unit->hp < uu_hp_to_repair(u, u2)) return A_ANY_ERROR;
    if (!has_enough_acp(unit, uu_acp_to_repair(u, u2))) return A_ANY_NO_ACP;
    for_all_material_types(m) {
	if (unit->supply[m] < um_to_repair(u, m)) return A_ANY_ERROR;
	if (unit->supply[m] < um_consumed_by_repair(u2, m)) return A_ANY_ERROR;
	/* (should check unit2 supplies also?) */
    }
    return A_ANY_OK;
}

/* The disband action destroys a unit in an "orderly" fashion, and can be
   undertaken voluntarily. */

prep_disband_action(unit, unit2)
Unit *unit, *unit2;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
    unit->act->nextaction.type = A_DISBAND;
    unit->act->nextaction.args[0] = (long) unit2;
    return TRUE;
}

/* Basic disbanding action. */

do_disband_action(unit, unit2)
Unit *unit, *unit2;
{
    int u, u2, m, amt, disb;
    int rslt = check_disband_action(unit, unit2);

    if (!valid(rslt)) return rslt;
    u2 = unit2->type;
    /* Recover some percentage of the unit's supply. */
    for_all_material_types(m) {
    	if (um_supply_per_disband(u2, m) > 0 && unit->supply[m] > 0) {
    	    amt = (unit->supply[m] * um_supply_per_disband(u2, m)) / 100;
    	    /* Unit always loses the amount, whether or not distributed. */
    	    unit->supply[m] -= amt;
    	    distribute_material(unit2, m, amt);
    	}
    }
    /* Remove hit points or kill the unit directly. */
    disb = u_hp_per_disband(u2);
    if (disb < unit2->hp) {
	unit2->hp -= disb;
    } else {
    	/* Pass around whatever we can get out of the unit itself. */
    	for_all_material_types(m) {
    	    if (um_recycleable(u2, m) > 0) {
    	    	distribute_material(unit2, m, um_recycleable(u2, m));
    	    }
    	}
    	/* should ensure vanish */
	kill_unit(unit2, DISBAND);
    }
    use_up_acp(unit, u_acp_to_disband(unit->type));
    return A_ANY_DONE;
}

/* Given a unit and a quantity of material, pass it out to nearby units. */

distribute_material(unit, m, amt)
Unit *unit;
int m, amt;
{
    Unit *unit2;

    /* Distribute to transport first. */
    if (amt > 0 && unit->transport != NULL) {
    	/* (should clip to capacity etc) */
    	unit->transport->supply[m] += amt;
    	amt = 0;
    }
    /* Then to any unit in the cell. */
    if (amt > 0) {
    }
    /* Then to any unit in an adjacent cell. */
    if (amt > 0) {
    }
    /* (should note anything that went to waste?) */
}

check_disband_action(unit, unit2)
Unit *unit, *unit2;
{
    if (!in_play(unit)) return A_ANY_ERROR;
    if (!in_play(unit2)) return A_ANY_ERROR;
    if (u_acp_to_disband(unit2->type) < 1) return A_ANY_CANNOT_DO;
    if (u_hp_per_disband(unit2->type) <= 0) return A_ANY_ERROR;
    if (!has_enough_acp(unit, u_acp_to_disband(unit2->type))) return A_ANY_NO_ACP;
    return A_ANY_OK;
}

/* Create a new unit that is similar to the original one, and give it
   some of the parts of the original unit. */
/* (New unit in same hex if possible or else in random adjacent hex.) */

prep_detach_action(unit, parts)
Unit *unit;
int parts;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
    unit->act->nextaction.type = A_DETACH;
    unit->act->nextaction.args[0] = parts;
    return TRUE;
}

do_detach_action(unit, parts)
Unit *unit;
int parts;
{
    int u = unit->type;
    Unit *unit2;
    int rslt = check_detach_action(unit, parts);

    if (!valid(rslt)) return rslt;
#if 0
    /* test if requested parts greater than original size */
    if ((unit2 = create_unit(u)) != NULL) {
	/* (fill in slots of new unit) */
    } else {
    }
#endif
    return A_ANY_DONE;
}

check_detach_action(unit, parts)
Unit *unit;
int parts;
{
    int u;

    if (!in_play(unit)) return A_ANY_ERROR;
    u = unit->type;
    if (u_parts(u) <= 1) {
	return A_ANY_ERROR;
    }
    if (!has_enough_acp(unit, 1)) return A_ANY_NO_ACP;
    return A_ANY_OK;
}

/* Merge two similar multi-part units into one larger unit. */

do_merge_action(unit, unit2)
Unit *unit, *unit2;
{
    int u, u2;
    int rslt = check_merge_action(unit, unit2);

    if (!valid(rslt)) return rslt;
    u = unit->type;
    u2 = unit2->type;
    /* Increase the unit2's hp by what's in this unit, and cap it. */
    unit2->hp += unit->hp;
    unit2->hp = max(unit2->hp, u_hp(u));
    /* Should affect morale etc according to proportions of mixing */
    /* Get rid of the unit that is acting (it was the one that expressed
       interest in being merged, after all). */
    /* (but more useful to do a sort of "pickup", sucking in other units
       as the unit moves along?) */
    kill_unit(unit, -1);  /* need a merge kill-reason? */
    return A_ANY_DONE;
}

check_merge_action(unit, unit2)
Unit *unit, *unit2;
{
    int u, u2;

    if (!in_play(unit) || unit2 == NULL) return A_ANY_ERROR;
    u = unit->type;
    u2 = unit2->type;
    if (u != u2) {
	return A_ANY_ERROR;
    }
    if (u_parts(u) <= 1) {
	return A_ANY_ERROR;
    }
    /* Units must be in same hex or adjacent. */
    if (distance(unit->x, unit->y, unit2->x, unit2->y) > 1) {
	return A_ANY_ERROR;
    }
    /* should check that unit doesn't become too big, or just
       suck up as many parts as possible or what? */
    if (!has_enough_acp(unit, 1)) return A_ANY_NO_ACP;
    return A_ANY_OK;
}

do_change_type_action(unit, nu, unit2)
Unit *unit, *unit2;
int nu;
{
    int rslt = check_change_type_action(unit, nu, unit2);

    if (!valid(rslt)) return rslt;
    return A_ANY_DONE;
}

check_change_type_action(unit, nu, unit2)
Unit *unit, *unit2;
int nu;
{
    if (!has_enough_acp(unit, 1)) return A_ANY_NO_ACP;
    return A_ANY_ERROR;
}

/* (what about "third-party transfers"?) */
/* (what about occs, garrisons, plans?) */

prep_change_side_action(unit, side, unit2)
Unit *unit, *unit2;
Side *side;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
    unit->act->nextaction.type = A_CHANGE_SIDE;
    unit->act->nextaction.args[0] = (long) side;
    unit->act->nextaction.args[1] = (long) unit2;
    return TRUE;
}

do_change_side_action(unit, side, unit2)
Unit *unit, *unit2;
Side *side;
{
    int rslt = check_change_side_action(unit, side, unit2);

    if (!valid(rslt)) return rslt;
    if (side_controls_unit(unit->side, unit2)) {
	/* If we own it, we can just change it. */
	unit_changes_side(unit2, side, -1, -1);
	rslt = A_ANY_DONE;
    } else {
	rslt = A_ANY_ERROR;
    }
    use_up_acp(unit, u_acp_to_change_side(unit->type));
    return rslt;
}

check_change_side_action(unit, side, unit2)
Unit *unit, *unit2;
Side *side;
{
    if (!in_play(unit)) return A_ANY_ERROR;
    if (!side_in_play(side)) return A_ANY_ERROR;
    if (!in_play(unit2)) return A_ANY_ERROR;
    if (!unit_allowed_on_side(unit2, side)) return A_ANY_ERROR;
    if (u_acp_to_change_side(unit->type) < 1) return A_ANY_CANNOT_DO;
    if (!has_enough_acp(unit, u_acp_to_change_side(unit->type)))
      return A_ANY_NO_ACP;
    return A_ANY_OK;
}

/* Change the terrain in the hex to something else. */

/* We don't need to ensure that the unit can exist on the new terrain
   type, because the designer is presumed to have set things up sensibly,
   because the unit might be in an appropriate transport, or because
   there is some actual use in such a bizarre shtick. */

/* What if engineers dig hole underneath enemy unit?  Should this be
   possible, or should there be a "can-dig-under-enemy" parm?  */

/* These should maybe check validity of new terr type for desired
   role (hex, bord, conn), or do in general period check. */

do_alter_cell_action(unit, x, y, t)
Unit *unit;
int x, y, t;
{
#if 0
    /* See if either terrain alteration is compatible with the
       desired terrain type. */
    if (ut_alter_hex_1(unit->type, terrain_at(x, y)) == t ||
	ut_alter_hex_2(unit->type, terrain_at(x, y)) == t) {
	/* Change the terrain to the new type. */
	set_terrain_at(x, y, t);
	/* Broadcast the change to anyone watching. */
	return TRUE;
    }
#endif
    return A_ANY_UNIMPL;
}

/* Change the terrain along a border to something else. */

do_add_terrain_action(unit, x, y, dir, t)
Unit *unit;
int x, y, dir, t;
{
#if 0
    int tx = unit->x, ty = unit->y;

    /* See if either terrain alteration is compatible with the
       desired terrain type. */
    /* (Should also check against both hex terr types for compat) */
    if (border_at(tx, ty, dir) == NOBORDER) {
	if (ut_add_border(unit->type, t)) {
	    set_border_at(tx, ty, dir, t);
	    return TRUE;
	}
    } else {
	if (ut_alter_border_1(unit->type, border_at(tx, ty, dir)) == t ||
	    ut_alter_border_2(unit->type, border_at(tx, ty, dir)) == t) {
	    /* Change the terrain to the new type. */
	    set_border_at(tx, ty, dir, t);
	    /* Broadcast the change to anyone watching. */
	    return TRUE;
	}
    }
#endif
    return A_ANY_UNIMPL;
}

/* Change the terrain of a connection to something else. */

do_remove_terrain_action(unit, x, y, dir, t)
Unit *unit;
int x, y, dir, t;
{
#if 0
    int tx = unit->x, ty = unit->y;

    return FALSE;

    /* See if either terrain alteration is compatible with the
       desired terrain type. */
    /* (Should also check against both hex terr types for compat) */
    if (connection_at(tx, ty, dir) == NOCONNECTION) {
	if (ut_add_connection(unit->type, t)) {
	    set_connection_at(tx, ty, dir, t);
	    return TRUE;
	}
    } else {
	if (ut_alter_connection_1(unit->type, connection_at(tx, ty, dir)) == t ||
	    ut_alter_connection_2(unit->type, connection_at(tx, ty, dir)) == t) {
	    /* Change the terrain to the new type. */
	    set_connection_at(tx, ty, dir, t);
	    /* Broadcast the change to anyone watching. */
	    return TRUE;
	}
    }
#endif
    return A_ANY_UNIMPL;
}

/* Execute a given action on a given unit. */

/* (assumes unit can act in the first place - valid?) */

execute_action(unit, action)
Unit *unit;
Action *action;
{
    int rslt = A_ANY_ERROR, n, (*fn)(), i;

    Dprintf("%s attempting %s with %d acp left",
	    unit_desig(unit), action_desig(action), unit->act->acp);

    if (!alive(unit) || !unit->act || unit->act->acp <= 0) return A_ANY_ERROR;

    n = strlen(actiondefns[(int) action->type].argtypes);
    fn = actiondefns[(int) action->type].dofn;
    /* Fill in "A" type args with the acting unit if they are NULL. */
    for (i = 0; i < n; ++i) {
	if (actiondefns[(int) action->type].argtypes[i] == 'A') {
	    if (action->args[i] == 0L) action->args[i] = (long) unit;
	}
    }
    switch (n) {
      case 0:
	rslt = (*fn)(unit);
	break;
      case 1:
	rslt = (*fn)(unit, action->args[0]);
	break;
      case 2:
	rslt = (*fn)(unit, action->args[0], action->args[1]);
	break;
      case 3:
	rslt = (*fn)(unit, action->args[0], action->args[1], action->args[2]);
	break;
      case 4:
	rslt = (*fn)(unit, action->args[0], action->args[1], action->args[2],
		     action->args[3]);
	break;
      default:
	case_panic("action arg count", n);
    }
    Dprintf("%s\n", hevtdefns[rslt].name);
    if (unit->side && side_has_ai(unit->side)) {
	ai_react_to_action_result(unit->side, unit, rslt);
    }
    if (unit->side && side_has_display(unit->side)) {
	update_action_result_display(unit->side, unit, rslt);
    }
    /* Check any scorekeepers that run after each action. */
    check_after_action_scores();
    /* Return success/failure so caller can use. */
    return rslt;
}

/* Basic check that unit has sufficient acp to do an action. */

has_enough_acp(unit, acp)
Unit *unit;
int acp;
{
    return ((unit->act->acp - acp) >= u_acp_min(unit->type));
}

/* This is true iff the unit has enough of each sort of supply to act. */

has_supply_to_act(unit)
Unit *unit;
{
    int m;

    for_all_material_types(m) {
	if (unit->supply[m] < um_to_act(unit->type, m)) return FALSE;
    }
    return TRUE;
}

/* Make the consumed acp disappear. */

use_up_acp(unit, acp)
Unit *unit;
int acp;
{
    /* This can sometimes be called on dead or non-acting units,
       so check first. */
    if (alive(unit) && unit->act) {
	unit->act->acp -= acp;
	/* Clue in the unit's display. */
	update_unit_display(unit->side, unit);
    }
}

/* Functions returning general abilities of a unit. */

can_create(unit)
Unit *unit;
{
    int u = unit->type, u2;
	
    for_all_unit_types(u2) {
	if (uu_acp_to_create(u, u2) > 0) return TRUE;
    }
    return FALSE;
}

can_complete(unit)
Unit *unit;
{
    int u = unit->type, u2;
	
    for_all_unit_types(u2) {
	if (uu_acp_to_build(u, u2) > 0) return TRUE;
    }
    return FALSE;
}

/* This tests whether the given unit is capable of doing repair. */

can_repair(unit)
Unit *unit;
{
    int u = unit->type, u2;
	
    for_all_unit_types(u2) {
	if (uu_acp_to_repair(u, u2) > 0) return TRUE;
    }
    return FALSE;
}

can_attack(unit)
Unit *unit;
{
    int u = unit->type, u2;
	
    for_all_unit_types(u2) {
	if (uu_hit(u, u2) > 0 && uu_damage(u, u2) > 0) return TRUE;
    }
    return FALSE;
}

can_fire(unit)
Unit *unit;
{
    int u = unit->type, u2;
	
    for_all_unit_types(u2) {
	if (uu_hit(u, u2) > 0 && uu_damage(u, u2) > 0) return TRUE;
    }
    return FALSE;
}

can_detonate(unit)
Unit *unit;
{
    return (u_acp_to_detonate(unit->type) > 0);
}

can_disband_at_all(side, unit)
Side *side;
Unit *unit;
{
    return (side != NULL
    	    && alive(unit)
	    && side_controls_unit(side, unit)
	    && (u_acp_to_disband(unit->type) > 0
	        || !completed(unit)
	        || side->designer));
	    	
}

/* Compose a legible description of a given action. */

char *
action_desig(act)
Action *act;
{
    int i;
    char ch, *str = actiondesigbuf;

    if (act->type == A_NONE) return "[]";
    sprintf(str, "[%s", actiondefns[act->type].name);
    for (i = 0; i < strlen(actiondefns[act->type].argtypes); ++i) {
	ch = (actiondefns[act->type].argtypes)[i];
	switch (ch) {
	  case 'U':
	  case 'A':
	    sprintf(str+strlen(str), " \"%s\"", unit_desig(find_unit(act->args[i])));
	    break;
	  default:
	    sprintf(str+strlen(str), " %d", act->args[i]);
	}
    }
    sprintf(str+strlen(str), "]");
    return actiondesigbuf;
}
