/* Copyright (c) 1987-1993  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* This file manages combat actions specifically. */

/* Rules of combat: the attacker hits the defender ("other") unit and its
   occupants, but the damage does not take effect right away.  If counter
   attacks are possible in this period, the defender always does so, with
   the same odds.  If the defender dies, then the attacker moves into the
   hex.  If the attacker dies, nothing happens.  If both survive, then the
   attacker may attempt to capture the defender. */

#include "conq.h"

extern int occdeath[];

char *summarize_units();

/* Remember what the main units involved are, so display is handled relative */
/* to them and not to any occupants. */

Unit *amain, *omain;

/* Hits on main units saved up, hits on occupants happen immediately. */

int ahit, ohit;

/* ... but the data is saved anyway, for message generation. */

int occhits[MAXUTYPES], occkills[MAXUTYPES];

prep_attack_action(unit, unit2, n)
Unit *unit, *unit2;
int n;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
    unit->act->nextaction.type = A_ATTACK;
    unit->act->nextaction.args[0] = (long) unit2;
    unit->act->nextaction.args[1] = n;
    return TRUE;
}

/* This is an attack on a given unit at a given level of commitment. */

do_attack_action(unit, unit2, n)
Unit *unit, *unit2;
int n;
{
    int u, u2;
    int rslt = check_attack_action(unit, unit2, n);

    if (!valid(rslt)) return rslt;
    u = unit->type;  u2 = unit2->type;
    one_attack(unit, unit2);
    if (alive(unit)) use_up_acp(unit, uu_acp_to_attack(u, u2));
    /* The defender in an attack has to take time to defend itself. */
    if (alive(unit2)) use_up_acp(unit2, uu_acp_to_defend(u, u2));
    return A_ANY_DONE;
}

check_attack_action(unit, unit2, n)
Unit *unit, *unit2;
int n;
{
    int u, u2;

    if (!in_play(unit) || !in_play(unit2) || n == 0) return A_ANY_ERROR;
    if (!has_enough_acp(unit, 1)) return A_ANY_NO_ACP;
    /* (should test if unit2 really present) */
    u = unit->type;
    u2 = unit2->type;
    if (uu_hit(u, u2) <= 0) return A_ANY_ERROR;
    /* Allow attacks if zero damage, this amounts to "harassment". */
    return A_ANY_OK;
}

prep_overrun_action(unit, x, y, z, n)
Unit *unit;
int x, y, z, n;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
    unit->act->nextaction.type = A_OVERRUN;
    unit->act->nextaction.args[0] = x;
    unit->act->nextaction.args[1] = y;
    unit->act->nextaction.args[2] = z;
    unit->act->nextaction.args[3] = n;
    return TRUE;
}

/* This is an attack on a given unit at a given level of commitment. */

do_overrun_action(unit, x, y, z, n)
Unit *unit;
int x, y, z, n;
{
    int u, u2, cleared = TRUE;
    Unit *unit2;
    int rslt = check_overrun_action(unit, x, y, z, n);

    if (!valid(rslt)) return rslt;
    u = unit->type;
    for_all_stack(x, y, unit2) {
	u2 = unit2->type;
	one_attack(unit, unit2);
	if (alive(unit)) use_up_acp(unit, uu_acp_to_attack(u, u2));
	/* The target of an attack has to take time to defend itself. */
	if (alive(unit2)) {
	    use_up_acp(unit2, uu_acp_to_defend(u, u2));
	    cleared = FALSE;
	}
    }
    if (cleared) {
	/* Enter the cleared cell now. */
	move_unit(unit, x, y);
	rslt = A_OVERRUN_SUCCEEDED;
    } else {
	rslt = A_OVERRUN_FAILED;
    }
    return rslt;
}

check_overrun_action(unit, x, y, z, n)
Unit *unit;
int x, y, z, n;
{
    int u;
    Unit *unit2;

    if (!in_play(unit) || !inside_area(x, y) || n == 0) return A_ANY_ERROR;
    if (!has_enough_acp(unit, 1)) return A_ANY_NO_ACP;
    u = unit->type;
    for_all_stack(x, y, unit2) {
	/* (should test if units here can be attacked en masse) */
    }
    return A_ANY_OK;
}

/* Return true if the attacker defeated the defender, and can therefore
   try to move into the defender's old position. */

one_attack(atker, unit2)
Unit *atker, *unit2;
{
    int u, ax = atker->x, ay = atker->y, ox = unit2->x, oy = unit2->y;
    Side *as = atker->side, *os = unit2->side;

    amain = atker;  omain = unit2;
    ahit = ohit = 0;
    for_all_unit_types(u) occhits[u] = occkills[u] = 0;
    attack_unit(atker, unit2);
    /* (this test should account for counterattacking occupants too,
       handle them in here somehow) */
/*    if (uu_counter_strength(unit2->type, atker->type) > 0) {
	attack_unit(unit2, atker);
    } */
    reckon_damage();
    see_exact(as, ax, ay);
    see_exact(as, ox, oy);
    see_exact(os, ax, ay);
    see_exact(os, ox, oy);
    update_cell_display(as, ax, ay, TRUE);
    update_cell_display(as, ox, oy, TRUE);
    update_cell_display(os, ax, ay, TRUE);
    update_cell_display(os, ox, oy, TRUE);
    all_see_hex(ax, ay);
    all_see_hex(ox, oy);
    attempt_to_capture_unit(atker, unit2);
    return (alive(atker) && unit_at(ox, oy) == NULL);
}

/* Shooting at a given location. */

prep_fire_at_action(unit, unit2, m)
Unit *unit, *unit2;
int m;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
    unit->act->nextaction.type = A_FIRE_AT;
    unit->act->nextaction.args[0] = (long) unit2;
    unit->act->nextaction.args[1] = m;
    return TRUE;
}

/* One can always shoot, if the hex is visible, but there might not
   not be anything to hit!  No counterattacks when shooting, and the
   results might not be visible to the shooter. */

do_fire_at_action(unit, unit2, r)
Unit *unit, *unit2;
int r;
{
    int ux = unit->x, uy = unit->y, uz = unit->z, ox, oy, oz, u;
    Unit *other;
    int rslt = check_fire_at_action(unit, unit2, r);

    if (!valid(rslt)) return rslt;
    update_fire_at_display(unit->side, unit, unit2, r, TRUE);
    update_fire_at_display(unit2->side, unit, unit2, r, TRUE);
    ox = unit2->x;  oy = unit2->y;  oz = unit2->z;
    amain = unit;  omain = unit2;
    ahit = ohit = 0;
    for_all_unit_types(u) occhits[u] = occkills[u] = 0;
    attack_unit(unit, unit2);
    reckon_damage();
    if (alive(unit)) use_up_acp(unit, u_acp_to_fire(unit->type));
    /*	if (alive(unit2)) use_up_acp(unit2, 1); */
    /* Each side sees what happened to its own unit. */
    update_unit_display(unit->side, unit);
    update_unit_display(unit2->side, unit2);
    /* The attacking side also sees the remote cell. */
    update_cell_display(unit->side, ox, oy, TRUE);
    update_cell_display(unit2->side, ox, oy, TRUE);
    /* Victim might see something in attacker's cell. */
    update_cell_display(unit2->side, ux, uy, TRUE);
    /* Actually, everybody might be seeing the combat. */
    all_see_hex(ux, uy);
    all_see_hex(ox, oy);
    /* Always expend the ammo (but only if r is a valid rtype). */
    return A_ANY_DONE;
}

/* Test a shoot action for plausibility. */

check_fire_at_action(unit, unit2, m)
Unit *unit, *unit2;
int m;
{
    int u, ux, uy, uz;

    if (!in_play(unit) || !in_play(unit2)) return A_ANY_ERROR;
    u = unit->type;
    ux = unit->x;  uy = unit->y;  uz = unit->z;
    if (!has_enough_acp(unit, u_acp_to_fire(u))) return A_ANY_NO_ACP;
    /* Check that target is in range. */
    if (!between(u_range_min(u),
		 distance(ux, uy, unit2->x, unit2->y),
		 u_range(u))) {
	return A_FIRE_AT_TOO_FAR;
    }
    /* Check elevations. */
    /* test if blocking terrain along the route */
    /* if elevations, test for clearance */
    /* test for enough of right kind of ammo */
    return A_ANY_OK;
}

/* Shooting at a given location. */

prep_fire_into_action(unit, x, y, z, m)
Unit *unit;
int x, y, z, m;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
    unit->act->nextaction.type = A_FIRE_INTO;
    unit->act->nextaction.args[0] = x;
    unit->act->nextaction.args[1] = y;
    unit->act->nextaction.args[2] = z;
    unit->act->nextaction.args[3] = m;
    return TRUE;
}

/* One can always shoot, if the hex is visible, but there might not
   not be anything to hit!  No counterattacks when shooting, and the
   results might not be visible to the shooter. */

do_fire_into_action(unit, x, y, z, r)
Unit *unit;
int x, y, z, r;
{
    int ux = unit->x, uy = unit->y, uz = unit->z, ox, oy, oz, u;
    Unit *other;
    int rslt = check_fire_into_action(unit, x, y, z, r);

    if (!valid(rslt)) return rslt;
    /* If any units at target, hit them. */
    for_all_stack(x, y, other) {
	ox = other->x;  oy = other->y;  oz = other->z;
	amain = unit;  omain = other;
	ahit = ohit = 0;
	for_all_unit_types(u) occhits[u] = occkills[u] = 0;
	attack_unit(unit, other);
	reckon_damage();
	/* Each side sees what happened to its own unit. */
	update_unit_display(unit->side, unit);
	update_unit_display(other->side, other);
	/* The attacking side also sees the remote cell. */
	update_cell_display(unit->side, ox, oy, TRUE);
	update_cell_display(other->side, ox, oy, TRUE);
	/* Victim might see something in attacker's cell. */
	update_cell_display(other->side, ux, uy, TRUE);
	/* Actually, everybody might be seeing the combat. */
	all_see_hex(ux, uy);
	all_see_hex(ox, oy);
	/* don't take moves though! */
    }
    /* Always expend the ammo (but only if r is a valid rtype). */
    /* We're always successful, even though the bombardment may have
       had little or no actual effect. */
    return A_ANY_DONE;
}

/* Test a shoot action for plausibility. */

check_fire_into_action(unit, x, y, z, m)
Unit *unit;
int x, y, z, m;
{
    int ux, uy, uz;

    if (!in_play(unit)) return A_ANY_ERROR;
    ux = unit->x;  uy = unit->y;  uz = unit->z;
    /* Check that target location is meaningful. */
    if (!inside_area(x, y)) {
	return A_FIRE_INTO_OUTSIDE_WORLD;
    }
    if (!has_enough_acp(unit, 1)) return A_ANY_NO_ACP;
    /* Check that target is in range. */
    if (!between(u_range_min(unit->type),
		 distance(ux, uy, x, y),
		 u_range(unit->type))) {
	return A_FIRE_INTO_TOO_FAR;
    }
    /* Check elevations. */
    /* test if blocking terrain along the route */
    /* if elevations, test for clearance */
    /* test for enough of right kind of ammo */
    return A_ANY_OK;
}

/* Test to see if enough ammo is available to make the attack. */
/* Need enough of *all* types - semi-bogus but too complicated otherwise? */

enough_ammo(unit, other)
Unit *unit, *other;
{
    int r;

    for_all_material_types(r) {
	if (um_hitby(other->type, r) > 0 &&
	    unit->supply[r] < um_hitswith(unit->type, r))
	  return FALSE;
    }
    return TRUE;
}

/* Single attack, no counterattack.  Check and use ammo - usage independent */
/* of outcome, but types used depend on unit types involved. */

attack_unit(atker, other)
Unit *atker, *other;
{
    int r;

    /* this should be an event needing a response */
/*    wake_unit(other, TRUE, WAKEENEMY, atker);  */
    if (alive(atker) && alive(other)) {
	if (enough_ammo(atker, other)) {
	    hit_unit(atker, other);
	    for_all_material_types(r) {
		if (um_hitby(other->type, r) > 0) {
		    atker->supply[r] -= um_hitswith(atker->type, r);
		}
	    }
	    /* Attack take same amount of time no matter what */
	    /* (should also consume actions?) */
/*	    atker->movesleft -= u_hittime(atker->type);  */
	}
    }
}

/* Make a single hit and maybe hit some passengers also.  Power of hit */
/* is constant, but chance is affected by neutrality, terrain, quality, */
/* and occupants' protective abilities.  If a hit is successful, it may */
/* have consequences on the defender's occupants, but limited by the */
/* protection that the transport provides. */

hit_unit(atker, other)
Unit *atker, *other;
{
    int chance, terr, hit = 0, a = atker->type, o = other->type;
    Unit *occ;
    Side *as = atker->side;

    chance = uu_hit(a,o);
/*    if (indep(atker)) chance += u_neutrality(a);  */
    terr = terrain_at(other->x, other->y);
/*    chance -= (chance * ut_defense(o, terr)) / 100; */
#if 0
    if (u_maxquality(a) > 0) {
	chance += ((chance * atker->quality * u_skillf(a)) /
		   u_maxquality(a)) / 100;
    }
    if (u_maxquality(o) > 0) {
	chance -= ((chance * other->quality * u_disciplinef(o)) /
		   u_maxquality(o)) / 100;
    }
#endif
    for_all_occupants(other, occ) {
	chance -= (chance * uu_protect(occ->type, o)) / 100;
    }
    if (probability(chance)) hit = uu_damage(a, o);
    if (as != NULL) {
#if 0
	side_atkstats(as, a, o)++;
	side_hitstats(as, a, o) += hit;
#endif
    }
    if (hit_unit_aux(atker, other, hit)) {
	for_all_occupants(other, occ) {
	    if (probability(100 - uu_protect(o, occ->type))) {
		hit_unit(atker, occ);
	    }
	}
    }
}

/* Do the hit itself.  Occupants are always hit/killed immediately, while
   the "main units" of the hexes don't get it till later.  In any case,
   messages are delayed.  Return true if the victim was hit but not killed. */

hit_unit_aux(atker, other, hit)
Unit *atker, *other;
int hit;
{
    int o = other->type, a = atker->type, chance, i;
    Side *as = atker->side, *os = other->side;

#if 0
    if (hit > 0 && mobile(o)) {
	chance = u_retreat(o);
	/* should adjust chance by morale etc */
	if (probability(chance)) {
	    if (retreat_unit(other)) {
		notify(as, "%s runs away!", unit_handle(as, other));
		notify(os, "%s runs away!", unit_handle(os, other));
		hit = 0;
	    }
	}
    }
#endif
    if (other == omain) {
	ohit = hit;
    } else if (other == amain) {
	ahit = hit;
    } else {
	if (hit >= other->hp) {
	    occkills[o]++;
	    kill_unit(other, COMBAT);
	    for_all_unit_types(i) occkills[i] += occdeath[i];
	} else if (hit > 0) {
	    occhits[o]++;
	    other->hp -= hit;
	}
    }
    return (alive(other) && hit > 0);
}

/* Hits on the main units have to be done later, so that mutual
   destruction works properly.  This function also does all the notifying. */

/* (Only the main units of a hex rate messages, occupants' fates are */
/* summarized briefly.) */

/* (What if occupants change type when killed, but transport vanishes?) */

reckon_damage()
{
    int o = omain->type, a = amain->type, i;
    Side *as = amain->side, *os = omain->side;
	char hitbuf[BUFSIZE], killbuf[BUFSIZE];

/*    strcpy(aabuf, unit_handle(as, amain));
    strcpy(aobuf, unit_handle(as, omain));
    strcpy(oabuf, unit_handle(os, amain));
    strcpy(oobuf, unit_handle(os, omain)); */

    if (ahit > 0) draw_blast(amain, omain->side, ahit);
    if (ohit > 0) draw_blast(omain, amain->side, ohit);
    if (ohit >= omain->hp) {
/*	notify(as, "%s %s %s!", aabuf, u_destroymsg(o), aobuf);
	notify(os, "%s %s %s!", oabuf, u_destroymsg(o), oobuf);  */
	if (u_wrecked_type(o) == NONUTYPE) {
	    kill_unit(omain, COMBAT);
	} else {
	    change_unit_type(omain, u_wrecked_type(o));
	    /* Restore to default hp for the new type. */
	    omain->hp = u_hp(o);
	}
	for_all_unit_types(i) occkills[i] += occdeath[i];
    } else if (ohit > 0) {
/*	notify(as, "%s hits %s!", aabuf, aobuf);
	notify(os, "%s hits %s!", oabuf, oobuf); */
	omain->hp -= ohit;
    } else {
	/* messages about missing not too useful */
    }
    summarize_units(hitbuf, occhits);
    summarize_units(killbuf, occkills);
    if (strlen(hitbuf) > 0) {
	if (strlen(killbuf) > 0) {
/*	    notify(as, "   (Also hit%s, killed%s)", hitbuf, killbuf);
	    notify(os, "   (%s hurt, %s killed)", hitbuf, killbuf); */
	} else {
/*	    notify(as, "   (Also hit%s)", hitbuf);
	    notify(os, "   (%s hurt)", hitbuf); */
	}
    } else {
	if (strlen(killbuf) > 0) {
/*	    notify(as, "   (Also killed%s)", killbuf);
	    notify(os, "   (%s killed)", killbuf); */
	}
    }
    if (ahit >= amain->hp) {
/*	notify(as, "%s %s %s!", aobuf, u_wreckmsg(a), aabuf);
	notify(os, "%s %s %s!", oobuf, u_wreckmsg(a), oabuf); */
	if (u_wrecked_type(a) == NONUTYPE) {
	    kill_unit(amain, COMBAT);
	} else {
	    change_unit_type(amain, u_wrecked_type(a));
	    /* Restore to default hp for the new type. */
	    amain->hp = u_hp(a);
	}
    } else if (ahit > 0) {
/*	notify(as, "%s hits %s!", aobuf, aabuf);
	notify(os, "%s hits %s!", oobuf, oabuf); */
	amain->hp -= ahit;
    } else {
	/* messages about missing not too useful */
    }
/*    if (u_selfdestruct(a)) kill_unit(amain, COMBAT);
    if (u_selfdestruct(o)) kill_unit(omain, COMBAT);  */
}

/* Retreat is a special kind of movement. */
/* Veterans should get several tries at retreating to a good place, perhaps */
/* one try per point of "veteranness"? */

retreat_unit(unit)
Unit *unit;
{
#if 0
    int dir;
    int success;

    dir = random_dir();
    success = move_delta(unit, dirx[dir], diry[dir]);
    return success;
#endif
}

prep_capture_action(unit, unit2)
Unit *unit, *unit2;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
    unit->act->nextaction.type = A_CAPTURE;
    unit->act->nextaction.args[0] = (long) unit2;
    return TRUE;
}

do_capture_action(unit, unit2)
Unit *unit, *unit2;
{
    int rslt = check_capture_action(unit, unit2);

    if (!valid(rslt)) return rslt;
    attempt_to_capture_unit(unit, unit2);
    return A_ANY_DONE;
}

check_capture_action(unit, unit2)
Unit *unit, *unit2;
{
    if (!in_play(unit) || !in_play(unit2)) return A_ANY_ERROR;
    if (uu_capture(unit->type, unit2->type) == 0) return A_ANY_ERROR;
    return A_ANY_OK;
}

/* Handle capture possibility and repulse/slaughter. */

/* The chance to capture an enemy is modified by several factors. */
/* Neutrals have a different chance to be captured, and presence of */
/* occupants should also has an effect.  Can't capture anything that is */
/* on a kind of terrain that the capturer can't go on, unless victim has */
/* "bridge effect". */

/* (Need a little better treatment of committed assaults, where lack of
   success == death.) */

attempt_to_capture_unit(atker, other)
Unit *atker, *other;
{
    int a = atker->type, o = other->type, chance;
    int ox = other->x, oy = other->y;
    Unit *occ;
    Side *as = atker->side, *os = other->side;

    if (alive(atker) && alive(other) && could_capture(a, o)) {
	if (impassable(atker, ox, oy) && !uu_bridge(o, a)) return;
	chance = uu_capture(a, o);
/*	if (u_maxquality(a) > 0) {
	    chance += ((chance * atker->quality * u_skillf(a)) /
		       u_maxquality(a)) / 100;
	} */
	for_all_occupants(other, occ) {
	    chance -= (chance * uu_protect(occ->type, o)) / 100;
	}
	if (probability(chance)) {
	    capture_unit(atker, other);
	} else if (atker->transport != NULL && 
		   (impassable(atker, ox, oy) ||
		    impassable(atker, atker->x, atker->y))) {
	    /* was the failed assault a commitment? */
/*	    notify(as, "Resistance... %s was slaughtered!",
		   unit_handle(as, atker));
	    notify(os, "Resistance... %s was slaughtered!",
		   unit_handle(os, atker));  */
	    kill_unit(atker, COMBAT);
	} else {
/*	    strcpy(aabuf, unit_handle(as, atker));
	    notify(as, "%s repulses %s!", unit_handle(as, other), aabuf);
	    strcpy(oabuf, unit_handle(os, atker));
	    notify(os, "%s repulses %s!", unit_handle(os, other), oabuf); */
	}
	/* (needs more sophistication) */
/*	atker->movesleft -= u_hittime(a);  */
    }
}

/* There are many consequences of a unit being captured.
   If the capturer is needed as a garrison, unload its occupants first.
   (what if erstwhile occupants can't stay in this hex though?) */

capture_unit(unit, pris)
Unit *unit, *pris;
{
    int u = unit->type, u2, px = pris->x, py = pris->y, i;
    int occs[MAXUTYPES], gains[MAXUTYPES], kills[MAXUTYPES];
    Unit *occ;
    Side *ps = pris->side, *us = unit->side;
	char hitbuf[BUFSIZE], killbuf[BUFSIZE];

    for_all_unit_types(u2) occs[u2] = gains[u2] = kills[u2] = 0;
/*    notify(us, "You captured %s!", unit_handle(us, pris));  */
/*    notify(ps, "%s has been captured by the %s!",
	   unit_handle(ps, pris), plural_form(us->name)); */
    /* (should alter or replace unit's plan completely) */
    /* Dispose of the occupants of our prisoner. */
    /* (should happen recursively) */
    /* (occupants might be able to escape also) */
    for_all_occupants(pris, occ) {
	occs[occ->type]++;
	if (/* u_change_side(occ->type) || */ could_capture(u, occ->type)) {
	    gains[occ->type]++;
	    /* side changing happens when whole unit changes */
	} else {
	    kills[occ->type]++;
	    kill_unit(occ, COMBAT);
	    for_all_unit_types(i) kills[i] += occdeath[i];
	}
    }
    summarize_units(hitbuf,  gains);
    summarize_units(killbuf, kills);
    summarize_units(spbuf, occs);
#if 0
    if (strlen(hitbuf) > 0) {
	if (strlen(killbuf) > 0) {
	    notify(us, "   (Also captured%s, killed%s)", hitbuf, killbuf);
	} else {
	    notify(us, "   (Also captured%s)", hitbuf);
	}
    } else if (strlen(killbuf) > 0) {
	notify(us, "   (Also killed%s)", killbuf);
    }
    if (strlen(spbuf) > 0) {
	notify(ps, "   (Lost%s)", spbuf);
    }
#endif
    /* The sad event itself */
    unit_changes_side(pris, us, CAPTURE, PRISONER);
    /* Possibly garrison the newly-captured unit. */
    if (uu_hp_to_garrison(u, pris->type) > 0) {
	/* needs to get more clever here, capturing unit might not enter */
	for_all_occupants(unit, occ) {
	    if (can_carry(pris, occ)) {
		leave_hex(occ);
		enter_hex(occ, px, py);
	    }
	}
	if (uu_hp_to_garrison(u, pris->type) >= unit->hp) {
	    /* This may kill some of the guard's occupants unnecessarily, but */
            /* I can't think of a better solution */
	    kill_unit(unit, GARRISON);
	} else {
	    unit->hp -= uu_hp_to_garrison(u, pris->type);
	}
    }
    see_exact(ps, px, py);
    update_cell_display(ps, px, py, TRUE);
    all_see_hex(px, py);
}

/* A detonate action just blasts the vicinity. */

prep_detonate_action(unit, x, y, z, unit2)
Unit *unit, *unit2;
int x, y, z;
{
    if (unit == NULL || unit->act == NULL) return FALSE;
    unit->act->nextaction.type = A_DETONATE;
    unit->act->nextaction.args[0] = x;
    unit->act->nextaction.args[1] = y;
    unit->act->nextaction.args[2] = z;
    unit->act->nextaction.args[3] = (long) unit2;
    return TRUE;
}

do_detonate_action(unit, x, y, z, unit2)
Unit *unit, *unit2;
int x, y, z;
{
    int rslt = check_detonate_action(unit, x, y, z, unit2);

    if (!valid(rslt)) return rslt;
    detonate_unit(unit2, x, y, z);
    if (alive(unit)) use_up_acp(unit, u_acp_to_detonate(unit->type));
    return A_ANY_DONE;
}

check_detonate_action(unit, x, y, z, unit2)
Unit *unit, *unit2;
int x, y, z;
{
	int u, u2;

    if (!in_play(unit) || !in_play(unit2)) return A_ANY_ERROR;
    u = unit->type;  u2 = unit2->type;
    /* The unit must actually be able to detonate. */
    if (u_acp_to_detonate(unit2->type) <= 0) return A_ANY_ERROR;
    if (!inside_area(x, y)) return A_ANY_ERROR;
    if (unit2 != unit) return A_ANY_ERROR; /* can only detonate self for now */
    /* Can only detonate in our own or an adjacent cell. */
    if (distance(unit2->x, unit2->y, x, y) > 1) return A_ANY_ERROR;
    if (!has_enough_acp(unit, u_acp_to_detonate(u))) return A_ANY_ERROR;
    return A_ANY_OK;
}

/* Actual detonation may occur by explicit action or automatically; this
   routine makes the detonation effects happen, pyrotechnics and all. */

/* (need to accumulate a damage report a la regular combat) */
/* (detonation effect at large distances should be patchy) */
/* (let elevations screen effects?) */
/* (interleave visual effects here - hook to display routines) */
/* (game design should be able to ask for a screen flash as extra) */

detonate_unit(unit, x, y, z)
Unit *unit;
int x, y, z;
{
    int u = unit->type, dir, x1, y1;
    Unit *unit2;

    /* Hit the detonating unit first.  This works because only the
       detonating unit's original type will be needed subsequently. */
    clobber_unit(unit, u_hp_per_detonation(u));
    /* Hit units at ground zero. */
    for_all_stack(x, y, unit2) {
    	if (unit2 != unit) {
	    	clobber_unit(unit2, uu_detonation_damage_at(u, unit2->type));
	    }
    }
    /* Hit units in adjacent cells if this is defined. */
    if (1) {
	    for_all_directions(dir) {
	    	if (interior_point_in_dir(x, y, dir, &x1, &y1)) {
			    for_all_stack(x1, y1, unit2) {
				    clobber_unit(unit2, uu_detonation_damage_adj(u, unit2->type));
			    }
	    	}
	    }
	}
    /* (should) Hit units at a distance, possibly throughout the world. */
    /* (Alter terrain, terrain materials?) */
    return TRUE;
}

clobber_unit(unit, hit)
Unit *unit;
int hit;
{
    if (hit >= unit->hp) {
	if (probability(u_detonate_on_death(unit->type))) {
	    detonate_unit(unit, unit->x, unit->y, unit->z);
	}
	if (alive(unit)) {
	    if (u_wrecked_type(unit->type) == NONUTYPE) {
		kill_unit(unit, COMBAT);
	    } else {
		change_unit_type(unit, u_wrecked_type(unit->type));
		/* Restore to default hp for the new type. */
		/* (what about rest of slots??) */
		unit->hp = u_hp(unit);
	    }
	}
    } else if (hit > 0) {
	/* (should check detonation upon hit, but need hitting unit) */
	unit->hp -= hit;
    }
    update_unit_display(unit->side, unit);  /* at least */
}
