/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* This file is devoted to commands and other functions related specifically
   to help.  Almost all of the Xconq help code is here.  Our main operating
   principle is that all game design info should be available
   via the help system, down to the last minor parameter. */

#include "conq.h"

/* The first help node in the chain, usually the list of topics. */

HelpNode *firsthelpnode = NULL;

/* The last help node. */

HelpNode *lasthelpnode = NULL;

int describe_topics();
int describe_news();
int describe_game_design();
int describe_game_modules();
int describe_utype();
int describe_mtype();
int describe_ttype();
int describe_scorekeepers();

/* Create an empty help node. */

HelpNode *
create_help_node()
{
    HelpNode *node = (HelpNode *) xmalloc(sizeof(HelpNode));

    node->key = NULL;
    node->fn = NULL;
    node->arg = 0;
    node->text = NULL;
    node->prev = node->next = NULL;
    return node;
}

/* Add a help node after the given node. */

HelpNode *
add_help_node(key, fn, arg, prevnode)
char *key;
int (*fn)(), arg;
HelpNode *prevnode;
{
    HelpNode *node = create_help_node(), *nextnode;

    node->key = key;
    node->fn = fn;
    node->arg = arg;
    if (prevnode != NULL) {
	nextnode = prevnode->next;
    } else {
	prevnode = lasthelpnode;
	nextnode = firsthelpnode;
    }
    node->prev = lasthelpnode;
    node->next = firsthelpnode;
    lasthelpnode->next = node;
    firsthelpnode->prev = node;
    lasthelpnode = node;
    return node;
}

/* Create a raw list of help topics by just iterating through all the nodes. */

describe_topics(arg, key, buf)
int arg;
char *key, *buf;
{
    HelpNode *tmp;

    for (tmp = firsthelpnode->next; tmp != firsthelpnode; tmp = tmp->next) {
	sprintf(buf+strlen(buf), "%s\n", tmp->key);
    }
}

/* Get the news file and put it into text buffer. */

describe_news(arg, key, buf)
int arg;
char *key, *buf;
{
    FILE *fp;

    if ((fp = fopen(newsfile_name(), "r")) != NULL) {
	sprintf(buf+strlen(buf), "XCONQ NEWS\n\n");
	while (fgets(spbuf, BUFSIZE-1, fp) != NULL) {
	    sprintf(buf+strlen(buf), "%s", spbuf);
	}
	fclose(fp);
    } else {
	sprintf(buf+strlen(buf), "(no news)");
    }
}

/* Create the initial help nodes, which is the index of topics, and link
   it to itself.  Subsequent nodes will be inserted later, after a game
   has been loaded. */

init_help()
{
    firsthelpnode = create_help_node();
    firsthelpnode->key = "topics";
    firsthelpnode->fn = describe_topics;
    firsthelpnode->prev = firsthelpnode->next = firsthelpnode;
    lasthelpnode = firsthelpnode;
    add_help_node("news", describe_news, 0, NULL);
}

/* This function creates the actual set of help nodes for the kernel. */

void
create_game_help_nodes()
{
    int u, m, t;

    add_help_node("current game design", describe_game_design, 0, NULL);
    add_help_node("current modules", describe_game_modules, 0, NULL);
    add_help_node("current scorekeepers", describe_scorekeepers, 0, NULL);
    for_all_unit_types(u) {
	add_help_node(u_type_name(u), describe_utype, u, NULL);
    }
    for_all_material_types(m) {
	add_help_node(m_type_name(m), describe_mtype, m, NULL);
    }
    for_all_terrain_types(t) {
	add_help_node(t_type_name(t), describe_ttype, t, NULL);
    }
    /* Invalidate any existing topics node. */
    firsthelpnode->text = NULL;
}

/* Return the string containing the text of the help node, possibly
   computing it first. */

char *
get_help_text(node)
HelpNode *node;
{
    if (node != NULL) {
	/* Maybe calculate the text to display. */
	if (node->text == NULL) {
	    if (node->fn != NULL) {
		if ((node->text = (char *) xmalloc(5000)) != NULL) {
		    (node->text)[0] = '\0';
		    node->textend = 0;
		    node->textsize = 5000;
		    (*(node->fn))(node->arg, node->key, node->text);
		    node->textend = strlen(node->text);
		} else {
		}
	    } else {
		sprintf(spbuf, "%s: No info available.", node->key);
		node->text = copy_string(spbuf);
		node->textend = strlen(node->text);
	    }
	}
	return node->text;
    } else {
	return NULL;
    }
}

/* Spit out all the general game_design parameters in a readable fashion.
   Some interfaces might have a better way of doing this, but that's up
   to the individual interface. */

describe_game_design(arg, key, buf)
int arg;
char *key, *buf;
{
    int u, r, t, i;

    strcat(buf, "\n");
    sprintf(spbuf, "This game design includes:\n");
    strcat(buf, spbuf);
    sprintf(spbuf,
	    "  %d unit types\n   %d material types\n   %d terrain types\n",
	    numutypes, nummtypes, numttypes);
    strcat(buf, spbuf);
    strcat(buf, "\n");
#if 0
    /* Should only do if country placing in effect... */
    sprintf(buf+strlen(buf),
	    "Countries are %d cells across, between %d and %d cells apart.\n",
	    2*g_min_radius()+1,
	    g_min_separation(), g_max_separation());
    if (g_see_all()) {
	sprintf(spbuf, "All units are always seen by all sides.\n");
	strcat(buf, spbuf);
    }
    strcat(buf, "\n");
    if (g_rt_for_game_max() > 0) {
	sprintf(spbuf, "Entire game must be over in %d minutes.\n",
		g_rt_for_game_max() / 60);
	strcat(buf, spbuf);
    }
    if (g_rt_per_side_max() > 0) {
	sprintf(spbuf, "Each side gets %d minutes of real time to act.\n",
		g_rt_per_side_max() / 60);
	strcat(buf, spbuf);
    }
#if 0
    for_all_terrain_types(t) {
	sprintf(buf+strlen(buf), "Terrain:  %s\n",
		t_type_name(t),
#if 0
		t_color(t),
		t_minalt(t), t_maxalt(t), t_minwet(t), t_maxwet(t)
#endif
		);
    }
    strcat(buf, "\n");
    for_all_material_types(r) {
	sprintf(buf+strlen(buf), "Material: %s\n",
		m_type_name(r));
    }
    strcat(buf, "\n");
    for_all_unit_types(u) {
	sprintf(buf+strlen(buf), "Unit:     %s  %s (%s)\n",
		utype_name_n(u, 1), u_type_name(u), u_help(u));
    }
    strcat(buf, "\n");
#endif
#ifndef SPECIAL
#if 0
    if (g_notes != NULL) {
	for (i = 0; g_notes[i] != NULL; ++i) {
	    sprintf(buf+strlen(buf), "%s\n", g_notes[i]);
	}
    }
#endif
#endif
#endif
}

/* Full details on the given type of unit.  This may be used either for */
/* online help or for building a descriptive file.  The icon will only */
/* show up for online help, otherwise the display calls have no effect. */

/* (The defaults should come from the *.def defaults!!) */

describe_utype(u, key, buf)
int u;
char *key, *buf;
{
    int r, t, u2;
    Obj *notes, *note;
    char *notestr;

	if (0 /* abbrevs to mention */) {
    	/* should display all the type names... */
    }
    /* The help string should be a sentence. */
    sprintf(buf+strlen(buf), "%s\n", u_help(u));
#if 0
    if (u_point_value(u) != 0) {
    	sprintf(buf+strlen(buf), "     (point value %d)\n", u_point_value(u));
    }
#endif
    sprintf(buf+strlen(buf), "Gets %d action points each turn.\n", u_acp(u));
    sprintf(buf+strlen(buf), "Max speed is %d cells/turn.\n", u_speed(u));
    sprintf(buf+strlen(buf), "Has %d hit points.\n", u_hp(u));
#if 0
#if 0
    if (u_seerange(u) == 1) {
	sprintf(buf+strlen(buf), "Chance to see others %d%%.\n", u_seebest(u));
    } else {
	sprintf(buf+strlen(buf), "Chance to see %d%% at 1 hex, to %d%% at %d hexes.\n",
		u_seebest(u), u_seeworst(u), u_seerange(u));
    }
    sprintf(buf+strlen(buf), "Own visibility is %d.\n", u_visibility(u));
#endif
/*    sprintf(buf+strlen(buf), "%s\n",
	    (u_change_side(u) ? "Can be made to change sides." : ""));  */
    sprintf(buf+strlen(buf), "%s\n",
	    (u_hp_per_disband(u) > 0 ? "Can be disbanded and sent home." : ""));
    sprintf(buf+strlen(buf), "\n");
    sprintf(buf+strlen(buf), "%s\n",
	    "  Material   ToBui  Prod Store  Eats ToMov  Hits HitBy\n");
    sprintf(buf+strlen(buf), "%s\n",
	    "               (0)   (0)   (0)   (0)   (0)   (0)   (0)");
    for_all_material_types(r) {
	sprintf(spbuf, "%10s: ", m_type_name(r));
	append_number(spbuf, um_tomake(u, r), 0);
	append_number(spbuf, um_produce(u, r), 0);
	append_number(spbuf, um_storage(u, r), 0);
	append_number(spbuf, um_consume(u, r), 0);
	append_number(spbuf, um_tomove(u, r), 0);
	append_number(spbuf, um_hitswith(u, r), 0);
	append_number(spbuf, um_hitby(u, r), 0);
	sprintf(buf+strlen(buf), "%s\n", spbuf);
    }
    sprintf(buf+strlen(buf), "\n");
    sprintf(buf+strlen(buf), "%s\n",
	    "   Terrain    Enter Leave Rand% Hide% Defn% Prod% Attr% Acdn%\n");
    sprintf(buf+strlen(buf), "%s\n",
	    "               (0)   (0)   (0)   (0)   (0)   (0)   (0)   (0)\n");
    for_all_terrain_types(t) {
	sprintf(spbuf, "%10s: ", t_type_name(t));
/*	append_number(spbuf, ut_enter_penalty(u, t), 0);  */
	append_number(spbuf, ut_leave_penalty(u, t), 0);
	append_number(spbuf, ut_randommove(u, t), 0);
	append_number(spbuf, ut_visibility(u, t), 0);
/*	append_number(spbuf, ut_defense(u, t), 0);  */
	append_number(spbuf, ut_productivity(u, t), 0);
#ifdef ATTRITION
	append_number(spbuf, ut_attrition(u, t), 0);
#else
	append_number(spbuf, 0, 0);
#endif
	append_number(spbuf, ut_accident(u, t), 0);
	sprintf(buf+strlen(buf), "%s\n", spbuf);
    }
    sprintf(buf+strlen(buf), "\n");
    sprintf(buf+strlen(buf), "%s%s\n",
	    "     Hit%  Damg  Cap% Guard  Pro%",
	    "  Holds Embrk Disem Mob% Bridg Build   Fix");
    sprintf(buf+strlen(buf), "%s%s\n",
	    "      (0)   (0)   (0)   (0)   (0)",
	    "   (0)   (1)   (0) (100)   (0)   (0)   (0)");
    for_all_unit_types(u2) {
	sprintf(spbuf, "%s: ", utype_name_n(u2, 1));
	append_number(spbuf, uu_hit(u, u2), 0);
	append_number(spbuf, uu_damage(u, u2), 0);
	append_number(spbuf, uu_capture(u, u2), 0);
	append_number(spbuf, uu_hp_to_garrison(u, u2), 0);
	append_number(spbuf, uu_protect(u, u2), 0);
	append_number(spbuf, uu_capacity_x(u, u2), 0);
	append_number(spbuf, uu_embark_penalty(u, u2), 0);
	append_number(spbuf, uu_disembark_penalty(u, u2), 0);
	append_number(spbuf, uu_mobility(u, u2), 100);
	append_number(spbuf, uu_bridge(u, u2), 0);
/*	append_number(spbuf, uu_make(u,u2), 0);  */
	append_number(spbuf, uu_repair(u,u2), 0);
	sprintf(buf+strlen(buf), "%s\n", spbuf);
    }
    sprintf(buf+strlen(buf), "\n");
#endif
    /* Display the designer's notes for this type. */
    if (u_notes(u) != lispnil) {
	sprintf(buf+strlen(buf), "\nNotes:\n");
	if (stringp(u_notes(u))) {
	    append_note(note, buf);
	} else if (consp(u_notes(u))) {
	    for (notes = u_notes(u); notes != lispnil; notes = cdr(notes)) {
		append_note(car(notes), buf);
	    }
	}
    }
}

append_note(note, buf)
Obj *note;
char *buf;
{
    char *notestr;

    if (stringp(note)) {
	notestr = c_string(note);
	if (strlen(notestr) > 0) { 
	    strcat(buf, notestr);
	    strcat(buf, " ");
	} else {
	    strcat(buf, "\n");
	}
    }
}

/* A simple table-printing utility. Blanks out default values so they don't */
/* clutter the table. */

append_number(buf, value, dflt)
char *buf;
int value, dflt;
{
    if (value != dflt) {
	sprintf(tmpbuf, "%5d ", value);
	strcat(buf, tmpbuf);
    } else {
	strcat(buf, "      ");
    }
}

describe_mtype(m, key, buf)
int m;
char *key, *buf;
{
    strcpy(buf, "no material info yet");
}

describe_ttype(t, key, buf)
int t;
char *key, *buf;
{
    strcpy(buf, "no terrain info yet");
}

/* {any side, <name>} to {win,lose}: [at least xx <utype>], between %s and %s */
/* where %s is printable dates. */
/* or "before %s" if start == 0. */

describe_scorekeepers(arg, key, buf)
int arg;
char *key, *buf;
{
    int i, u, r, num;
    Scorekeeper *sk;
    char tmp1[1000], tmp2[1000];

    if (scorekeepers == NULL) {
	sprintf(buf+strlen(buf), "No scores are being kept.");
    } else {
	for_all_scorekeepers(sk) {
	    sprintf(buf+strlen(buf), "a scorekeeper");
	}
    }
}

#ifdef USE_CONSOLE

/* Print the news file onto the console if there is anything to print. */

print_any_news()
{
    FILE *fp;

    if ((fp = fopen(newsfile_name(), "r")) != NULL) {
	printf("                              XCONQ NEWS\n\n");
	while (fgets(spbuf, BUFSIZE-1, fp) != NULL) {
	    fputs(spbuf, stdout);
	}
	/* Add another blank line, to separate from init printouts. */
	printf("\n");
	fclose(fp);
    }
}

#endif /* USE_CONSOLE */

/* Generate a readable description of the game (design) being played. */
/* This works by writing out appropriate help nodes, along with some
   indexing material.  This does *not* do interface-specific help,
   such as commands. */

print_game_description_to_file(fp)
FILE *fp;
{
    HelpNode *node;

    /* (need to work on which nodes to dump out) */
    for (node = firsthelpnode; node != firsthelpnode; node = node->next) {
	get_help_text(node);
	if (node->text != NULL) {
	    fprintf(fp, "\014\n%s\n", node->key);
	    fprintf(fp, "%s\n", node->text);
	}
    }
}
